/**********************************************************************/
/***                                                                  */
/***        Author: Luis Eduardo Gamboa Guzmn                        */
/***        Last Updated: October 22th 2005                           */
/***                                                                  */
/***        Facultad de Ingeniera Elctrica                          */
/***        Universidad Michoacana de San Nicolas de Hidalgo          */
/***                                                                  */
/**********************************************************************/

#include <memory.h>
#include <stdlib.h>
#include <math.h>
#include "dsp.h"

/* Calcula la FFT de in (real, de tamao N), la parte real e imaginaria del resultado 
	se almacena en outr y outi respectivamente. Slo regresa N/2 coeficientes */
void fft(double *in,double *outr,double *outi,int N)
{
	int i;
	fftw_complex *out;
	fftw_plan plan;
	double *inp;

	inp=(double *)fftw_malloc(sizeof(double)*N);
	memcpy(inp,in,sizeof(double)*N);
	out=(fftw_complex *)fftw_malloc(sizeof(fftw_complex)*N);
	plan=fftw_plan_dft_r2c_1d(N, inp, out,FFTW_ESTIMATE);
	fftw_execute(plan);
	fftw_destroy_plan(plan);

	/*if(N%2)
	{
		for(i=0;i<N/2+1;i++)
		{
			outr[N-i-1]=out[i+1][0]/N;
			outi[N-i-1]=-out[i+1][1]/N;
			outr[i]=out[i][0]/N;
			outi[i]=out[i][1]/N;
		}
	}
	else
	{*/
		for(i=0;i<N/2;i++)
		{
			//outr[N-i-1]=out[i+1][0]/N;
			//outi[N-i-1]=-out[i+1][1]/N;
			outr[i]=out[i][0];//N;
			outi[i]=-out[i][1];//N;
		}
	//}
	fftw_free(out);
	fftw_free(inp);
}

/* Calcula la IFFT de inr e ini (real e imaginaria, tamao N), el resultado
	se almacena en outr, slo calcula la parte real */
void ifft(double *inr,double *ini,double *outr,int N)
{
	int i;
	fftw_complex *in;
	fftw_plan plan;

	in=(fftw_complex *)fftw_malloc(sizeof(fftw_complex)*N);
	for(i=0;i<N;i++)
	{
		in[i][0]=inr[i];
		in[i][1]=ini[i];
	}

	plan=fftw_plan_dft_c2r_1d(N,in,outr,FFTW_ESTIMATE|FFTW_DESTROY_INPUT);
	fftw_execute(plan);
	fftw_destroy_plan(plan);
	fftw_free(in);
}

void fftblock(double *x,double *window,double *fftr,double *ffti,int blocklength)
{
	double *block=(double *)malloc(sizeof(double)*blocklength);
	int i;

	for(i=0;i<blocklength;i++)
			block[i]=x[i]*window[i];

	fft(block,fftr,ffti,blocklength);
	free(block);
}

double *fftspl(double *xr,double *xi,int Nfreq)
{
	double *spl=(double *)malloc(sizeof(double)*Nfreq);
	double fg,mag2;
	int i;
	int N=Nfreq*2;

	fg=4.0/((double)(N*N)*windowgain);
	for(i=0;i<Nfreq;i++)
	{
		mag2=xr[i]*xr[i]+xi[i]*xi[i];
		spl[i]=96.0+10.0*log10(fg*mag2);
		if(spl[i]<0.0 || mag2==0.0)
			spl[i]=0.0;
	}

	return spl;
}

void mdct4(double *in,double *out,int N)
{
	int r,k,r2,k2;
	int N2,N4;
	double a,b,c,d,auxd,theta,*inrot;
	//double sqrtN=sqrt(N);
	fftw_complex *inc,*outc;
	fftw_plan plan;

	N2=N/2;
	N4=N/4;

	inc=(fftw_complex *)fftw_malloc(sizeof(fftw_complex)*N4);
	outc=(fftw_complex *)fftw_malloc(sizeof(fftw_complex)*N4);
	inrot=(double *)malloc(sizeof(double)*N);
	auxd=PI/N2;

	for(r=0;r<N4;r++)
		inrot[r] = -in[r+3*N4];
	for(;r<N;r++)
		inrot[r] = in[r-N4];

	for(r=0;r<N4;r++)
	{
		theta=auxd*(r+0.125);
		r2=2*r;
		a=0.5*(inrot[r2]-inrot[N-1-r2]);
		b=-0.5*(inrot[N2+r2]-inrot[N2-1-r2]);
		c=cos(theta);
		d=-sin(theta);
		inc[r][0]=a*c-b*d;
		inc[r][1]=a*d+b*c;
	}

	/* FFT N/4 */
	plan=fftw_plan_dft_1d(N4, inc, outc,FFTW_FORWARD,FFTW_ESTIMATE);
	fftw_execute(plan);
	fftw_destroy_plan(plan);

	for(k=0;k<N4;k++)
	{
		theta=auxd*(k+0.125);
		a=outc[k][0];
		b=outc[k][1];
		c=cos(theta);
		d=-sin(theta);

		k2=2*k;
		out[k2]=2.0*(a*c-b*d);// /sqrtN;
		out[N2-1-k2]=-2.0*(a*d+b*c);// /sqrtN;
	}
	fftw_free(inc);
	fftw_free(outc);
	free(inrot);
}

void imdct4(double *in,double *out,int K)
{
	fftw_plan plan;
	int k;
	int N=2*K;
	int K2=K/2,k2;
	double auxd,theta,a,b,c,d,*outrot;
	//double sqrtN=sqrt(N);

	fftw_complex *inc,*outc;
	inc=(fftw_complex *)fftw_malloc(sizeof(fftw_complex)*K2);
	outc=(fftw_complex *)fftw_malloc(sizeof(fftw_complex)*K2);
	outrot=(double *)malloc(sizeof(double)*N);
	
	auxd=PI/K;
	for(k=0;k<K2;k++)
	{
		theta=auxd*(k+0.125);
		k2=k*2;
		a=0.5*in[k2];
		b=0.5*in[K-1-k2];
		c=cos(theta);
		d=-sin(theta);
		inc[k][0]=a*c-b*d;
		inc[k][1]=a*d+b*c;
	}

	plan=fftw_plan_dft_1d(K2, inc, outc,FFTW_FORWARD,FFTW_ESTIMATE);
	fftw_execute(plan);
	fftw_destroy_plan(plan);

	for(k=0;k<K2;k++)
	{
		theta=auxd*(k+0.125);
		a=outc[k][0];
		b=outc[k][1];
		c=cos(theta);
		d=-sin(theta);
		outc[k][0]=8.0*(a*c-b*d)/(double)N;
		outc[k][1]=8.0*(a*d+b*c)/(double)N;
	}
	for(k=0;k<K2;k++)
	{
		k2=2*k;
		outrot[k2]=outc[k][0];
		outrot[K+k2]=outc[k][1];
	}
	for(k=1;k<N;k=k+2)
		outrot[k]=-outrot[N-1-k];

	for(k=0;k<3*K2;k++)
		out[k]=outrot[k+K2];//*sqrtN;
	for(;k<N;k++)
		out[k]=-outrot[k-3*K2];//*sqrtN;
	fftw_free(inc);
	fftw_free(outc);
	free(outrot);
}

void mdctblock(double *x,double *window,double *result,int blocklength)
{
	double *block=(double *)malloc(sizeof(double)*blocklength);
	int i;

	for(i=0;i<blocklength;i++)
			block[i]=x[i]*window[i];

	mdct4(block,result,blocklength);
	free(block);
}

void imdctblock(double *buffer,double *window,double *block,int blocklength)
{
	double *blocki=(double *)malloc(sizeof(double)*blocklength);
	int i;

	imdct4(block,blocki,blocklength/2);

	for(i=0;i<blocklength;i++)
		buffer[i]=blocki[i]*window[i];
	free(blocki);
}

double *mdctspl(double *x,int Nfreq)
{
	//double *xabs=(double *)malloc(sizeof(double)*Nfreq);
	double *spl=(double *)malloc(sizeof(double)*Nfreq);
	//double max=0.0;
	double fg;
	//double sqrtN2=2*N;
	int i;
	int N=Nfreq*2;

	// window gain is always 0.5 for valid MDCT windows
	fg=8.0/((double)(N*N)*0.5);
	for(i=0;i<Nfreq;i++)
	{
		spl[i]=96.0+10.0*log10(fg*x[i]*x[i]);//*sqrtN2);
		if(spl[i]<0.0 || x[i]==0.0)
			spl[i]=0.0;
	}

	return spl;
}

void sinewindow(double *window,int N)
{
	int i;
	for(i=0;i<N;i++)
		window[i]=sin(PI*((double)(i)+0.5)/N);
}

/* kbdwindow && BesselI0 son versiones modificadas de las funciones de 
Craig Stuart Sapp <craig@ccrma.stanford.edu>
http://ccrma.stanford.edu/courses/422/projects/kbd/kbdwindow.cpp
*/
void kbdwindow(double *window,int N,double alpha)
{
	double sumvalue = 0.0,aux;
	int i;
   
	for (i=0; i<N/2; i++)
	{
		aux = 4.0*i/N - 1.0;
		aux*=aux;
		sumvalue += BesselI0(PI * alpha * sqrt(1.0 - aux));
		window[i] = sumvalue;
	}

	// need to add one more value to the nomalization factor at size/2:
	aux = 4.0*(N/2)/N - 1.0;
	aux*=aux;
	sumvalue += BesselI0(PI * alpha * sqrt(1.0 - aux));

	// normalize the window and fill in the righthand side of the window:
	for (i=0; i<N/2; i++)
	{
		window[i] = sqrt(window[i]/sumvalue);
		window[N-1-i] = window[i];
	}
}
double BesselI0(double x)
{
	double denominator;
	double numerator;
	double z;

	if (x == 0.0)
	{
		return 1.0;
	}
	else
	{
		z = x * x;
		numerator = (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* (z* 
				(z* 0.210580722890567e-22  + 0.380715242345326e-19 ) +
				0.479440257548300e-16) + 0.435125971262668e-13 ) +
				0.300931127112960e-10) + 0.160224679395361e-7  ) +
				0.654858370096785e-5)  + 0.202591084143397e-2  ) +
				0.463076284721000e0)   + 0.754337328948189e2   ) +
				0.830792541809429e4)   + 0.571661130563785e6   ) +
				0.216415572361227e8)   + 0.356644482244025e9   ) +
				0.144048298227235e10);

		denominator = (z*(z*(z-0.307646912682801e4)+
				0.347626332405882e7)-0.144048298227235e10);
	}
	return -numerator/denominator;
}

void vorbiswindow(double *window,int N)
{
	int i;
	double aux;
	for(i=0;i<N;i++)
	{
		aux=sin(PI*((double)(i)+0.5)/N);
		aux*=aux;
		window[i]=sin(PI/2 * aux);
	}
}

void rectangularwindow(double *window,int N)
{
	int i;
	double aux=1.0/sqrt(2.0);
	for(i=0;i<N;i++)
		window[i]=aux;
}

void hanningwindow(double *window,int N)
{
	int i;
	double aux=2.0*PI/N;
	for(i=0;i<N;i++)
		window[i]=0.5*(1.0-cos(aux*(i+0.5)));
	windowgain=3.0/8.0;
}

int verifywindow(double *window,int N)
{
	int i,Ns2=N/2;
	double sum;
	for(i=0;i<Ns2;i++)
	{
		sum=window[i]*window[i] + window[Ns2+i]*window[Ns2+i];
		printf("%d %e\n",i,sum);
	}
	return 1;
}

double computewindowgain(double *window,int N)
{
	int i;
	double windowgain=0.0;
	for(i=0;i<N;i++)
		windowgain+=window[i]*window[i];
	windowgain/=(double)N;
	printf("%e\n",windowgain);
	return windowgain;
}

