/**********************************************************************/
/***                                                                  */
/***        Author: Luis Eduardo Gamboa Guzmn                        */
/***        Last Updated: June 6th 2005                               */
/***                                                                  */
/***        Facultad de Ingeniera Elctrica                          */
/***        Universidad Michoacana de San Nicolas de Hidalgo          */
/***                                                                  */
/**********************************************************************/

#include "psychoacoustics.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "dsp.h"

#define ALPHA 1

short int *splflags=NULL;
double *bark=NULL;
double *thresholdofhearing=NULL;
double *globalmask=NULL;
int barklimits[26][2];
short int barkflags[25];


void mpegpm2(double spl,int todo,int Nfreq,int start,int end,double delta,double alpha)
{
	double dz,aux1,aux2,aux3,aux4;
	double adding;
	int i;

	for(i=start;i<end;i++)
	{
		dz=bark[i]-bark[todo];

		aux1=1.05*dz+0.474;
		aux1*=aux1;
		aux2=1.05*dz-0.5;
		aux2*=aux2;
		aux3=aux2-2.0*(1.05*dz-0.5);
		if(aux3>0.0)
			aux3=0.0;
		aux4=15.8111389+7.5*(1.05*dz+0.474)-17.5*sqrt(1.0+aux1)+8.0*aux3+spl-delta;

		if(aux4>globalmask[i])
		{
			/* Mask addition */
			if(aux4-globalmask[i]<3.0)
				adding=6.0;
			else
				adding=0.0;
			globalmask[i]=aux4+adding;
		}
	}
}


void initpsychoacoustic(int Nfreq,int maxfreq)
{
	int i,j;
	double f;
	if(bark!=NULL)
		free(bark);
	if(splflags!=NULL)
		free(splflags);
	if(thresholdofhearing!=NULL)
		free(thresholdofhearing);
	if(globalmask!=NULL)
		free(globalmask);

	bark=(double *)malloc(sizeof(double)*Nfreq);
	splflags=(short int*)malloc(sizeof(short int)*Nfreq);
	thresholdofhearing=(double *)malloc(sizeof(double)*Nfreq);
	globalmask=(double *)malloc(sizeof(double)*Nfreq);
	
	/* Frequency to bark mapping and threshold of hearing */
	for(i=0;i<Nfreq;i++)
	{
	    f = (double)(i*maxfreq)/Nfreq;
		bark[i] = 13.0*atan(0.76*f/1000.0)+3.5*atan((f/7500.0)*(f/7500.0));

		f/=1000.0;
		thresholdofhearing[i]=3.64*pow(f,-0.8)-6.5*exp(-0.6*(f-3.3)*(f-3.3))+0.001*f*f*f*f;
		thresholdofhearing[i]/=2.0;
		/* Max TOH at 75dB */
		if(thresholdofhearing[i]>75.0)
			thresholdofhearing[i]=75.0;
	}
	
	/* Calculates bark limits */
	for(i=1,j=0;i<=25 && j<Nfreq;i++)
	{
		barklimits[i-1][0]=j;
		for(;bark[j]<i && j<Nfreq;j++);
		barklimits[i-1][1]=j-1;
		//printf("%d %d %d %d\n",i-1,barklimits[i-1][0],barklimits[i-1][1],barklimits[i-1][1]-barklimits[i-1][0]+1);
	}
	barklimits[25][0]=Nfreq;
	barklimits[25][1]=Nfreq;
}


void freepsychoacoustic(void)
{
	if(bark!=NULL)
		free(bark);
	if(splflags!=NULL)
		free(splflags);
	if(thresholdofhearing!=NULL)
		free(thresholdofhearing);
	if(globalmask!=NULL)
		free(globalmask);

	bark=NULL;
	splflags=NULL;
	thresholdofhearing=NULL;
	globalmask=NULL;
}

void computemask(double *fftr,double *ffti,int Nfreq,int maxfreq,double *mask)
{
	int i,todo;
	double *spl;
	double delta=12.0; /* Small masking */
	int	sub=4;
	int add=11;
	int start,end;


	//spl=mdctspl(mdct,Nfreq);
	spl=fftspl(fftr,ffti,Nfreq);

	for(i=0;i<Nfreq;i++)
	{
		splflags[i]=UNMASKED;
		globalmask[i]=0.0;
	}
	for(i=0;i<25;i++)
		barkflags[i]=UNMASKED;

	todo=nexttocomputemask(spl,splflags,Nfreq);
	while(todo!=-1)
	{
		start=(int)(bark[todo])-sub;
		if(start<0)
			start=0;
		else
			start=barklimits[start][0];

		end=(int)(bark[todo])+add;
		if(end>24)
			end=barklimits[24][1];
		else
			end=barklimits[end][1];

		mpegpm2(spl[todo],todo,Nfreq,start,end,delta,ALPHA);
		for(i=start;i<end;i++)
		{
			if(globalmask[i]>spl[i])
			//if((10.0*log10(globalmask[i])/ALPHA)>spl[i])
				splflags[i]=MASKED;
		}
		todo=nexttocomputemask(spl,splflags,Nfreq);
	}

	/* Post process global mask */
	for(i=0;i<Nfreq;i++)
	{
		if(thresholdofhearing[i]>globalmask[i])
			globalmask[i]=thresholdofhearing[i];
	}

	memcpy(mask,globalmask,sizeof(double)*Nfreq);

	free(spl);
}

int nexttocomputemask(double *spl,short int *splflags,int Nbark)
{
	int i,j=-1,k;
	double max;
	for(i=0;i<25;i++)
	{
		if(barkflags[i]<COMPUTED)
		{
			max=0.0;
			for(k=barklimits[i][0];k<=barklimits[i][1];k++)
			{
				if(splflags[k]==UNMASKED && spl[k]>max)
				{
					j=k;
					max=spl[k];
				}
			}
			barkflags[i]++;
			if(j!=-1)
			{
				splflags[j]=COMPUTED;
				return j;
			}
		}
	}
	return j;
}

double signaltomaskratio(double *spl,double *mask,int b)
{
	int i;
	double maxsignal,minmask;

	maxsignal=spl[barklimits[b][0]];
	minmask=mask[barklimits[b][0]];
	for(i=barklimits[b][0];i<=barklimits[b][1];i++)
	{
		if(mask[i]<minmask)
			minmask=mask[i];
		if(maxsignal<spl[i])
			maxsignal=spl[i];
	}

	return maxsignal-minmask;
}

int getbandsize(int b)
{
	return 1+barklimits[b][1]-barklimits[b][0];
}

void getbandlimits(int b,int *lower,int *upper)
{
	*lower=barklimits[b][0];
	*upper=barklimits[b][1];
}
